// Rotation convention:
// Origin (0,0) is top-left.
// Edges around a vertex are linked in counter-clockwise order.
// Hence the three vertices of each triangle are visited in clockwise order.

typedef double CoordT;

typedef struct EdgeT EdgeT;
typedef struct TriangleT TriangleT;

typedef struct {
  CoordT x, y;
  EdgeT *firstEdge;
  MagickBooleanType visited;
  double * pArgs; /* Pointer to arguments for this point. */
} PointT;

typedef struct EdgeT {
  PointT *org, *dst;
  EdgeT *orgNext, *dstNext;
  int numTri;       /* 1 or 2. 1 means this is on convex hull. */
  TriangleT * tri;  /* A triangle that has this as an edge. */
} EdgeT;

typedef struct TriangleT {
  PointT *p0, *p1, *p2;
  double denominator;
  CoordT centX, centY;
} TriangleT;

typedef enum {
  oExtrapolate,
  oClamp,
  oBackground
} OutsideT;

typedef struct {
  /* Caller should populate these. */
  MagickBooleanType verbose;
  int NumPoints;
  int NumArgsPp;  /* Number of arguments per point, >=2. */
  double * pArgs; /* Pointer to arguments for all points. */
  int precision;
  //MagickBooleanType background; // If true, don't bother calculating abc for pixels outside convex hull.
  OutsideT outside;
  double sigContrast, sigMidpoint;
  double power;
  MagickBooleanType CoordsLast;  // If true, for each point, coordinates come after the data, not before.

  /* Internal use. */
  int MaxNumPts;
  int NumPts;
  PointT * points;
  int MaxNumEdges;
  int NumEdges;
  EdgeT * edges;
  CoordT minX, minY, maxX, maxY;
  int MaxNumTris;
  int NumTris;
  TriangleT * tris;
} SparseTriT;

typedef struct {
  SparseTriT * st;
  int PrevFndTri;
} SparseTriParaT;


typedef struct {
  PointT *pa, *pb;
  CoordT SqrDist;
} CandidateEdgeT;

typedef enum {
  iNoIntersect,
  iSegment,
  iEndPointAB,
  iEndPointCD,
  iMidPoint
} IntersectT;

#define perp(u,v)  ((u).x * (v).y - (u).y * (v).x)  // perp product  (2D)

#define VectorT PointT

#define TRI_EPSILON 1e-7


static IntersectT SegsIntersect (PointT * A, PointT * B, PointT * C, PointT * D)
// Returns whether segments AB and CD intersect.
// Assumes A!=B and C!=D.
{
  assert (A != B);
  assert (C != D);

  VectorT AB, CD, CA;
  AB.x = B->x - A->x; //u
  AB.y = B->y - A->y;
  CD.x = D->x - C->x; //v
  CD.y = D->y - C->y;
  CA.x = A->x - C->x; //w
  CA.y = A->y - C->y;

  double Dp = perp(AB,CD);

  if (fabs(Dp) < TRI_EPSILON) {
    // They are parallel.
    if (perp(AB,CA) != 0 || perp(CD,CA) != 0) return iNoIntersect;
    // They are colinear.

    double t0, t1;
    VectorT CB;
    CB.x = B->x - C->x;
    CB.y = B->y - C->y;
    if (CD.x != 0) {
      t0 = CA.x / CD.x;
      t1 = CB.x / CD.x;
    } else {
      t0 = CA.y / CD.y;
      t1 = CB.y / CD.y;
    }
    if (t0 > t1) {
      double t = t0;
      t0 = t1;
      t1 = t;
    }
    if (t0 > 1 || t1 < 0) return iNoIntersect;
    t0 = (t0 < 0) ? 0 : t0;
    t1 = (t1 > 1) ? 1 : t1;
    //if (t0 == t1) {
    if (t1-t0 < TRI_EPSILON) {
      // They intersect at a point.
      // *I0 = S2.P0 + t0 * v
      if (t0 == 0 || t0 == 1) return iEndPointCD;
      return iMidPoint;
    }
    // They intersect at a segment.
    // *I0 = S2.P0 + t0 * v;
    // *I1 = S2.P0 + t1 * v;
    return iSegment;
  }

  // The segments are not parallel.

  double sI = perp(CD,CA) / Dp;

  if (sI < 0 || sI > 1) return iNoIntersect;

  double tI = perp(AB,CA) / Dp;

  if (tI < 0 || tI > 1) return iNoIntersect;

  // *I0 = S1.P0 + sI * u;
  if (sI == 0 || sI == 1) return iEndPointAB;
  if (tI == 0 || tI == 1) return iEndPointCD;
  return iMidPoint;
}

static MagickBooleanType IsHidden (PointT * A, PointT * B, PointT * C, PointT * D)
// Looking from A to B,
// returns whether B is hidden by edge from C to D.
{
  IntersectT it = SegsIntersect (A, B, C, D);

  return (it == iSegment || it == iMidPoint);
}

static void UnitVect (PointT * A, PointT * B, PointT *C)
// Given vector AB, sets C so AC is the unit vector along AB.
// Assumes A and B are distinct. Otherwise, the unit vector is undefined.
{
  CoordT dx = B->x - A->x;
  CoordT dy = B->y - A->y;

  CoordT len = sqrt (dx*dx + dy*dy);

  assert (len != 0);

  C->x = A->x + dx/len;
  C->y = A->y + dy/len;
}

static inline CoordT PtsDistSqr (PointT * A, PointT * B)
// Returns squared distance between A and B.
{
  CoordT dx = B->x - A->x;
  CoordT dy = B->y - A->y;

  return (dx*dx + dy*dy);
}

static inline CoordT PtsDist (PointT * A, PointT * B)
// Returns distance between A and B.
{
  return sqrt (PtsDistSqr (A, B));
}

static inline CoordT AreaX2 (PointT * A, PointT * B, PointT * C)
{
  return (- B->y * C->x
          + A->y * (-B->x + C->x)
          + A->x * (B->y - C->y)
          + B->x * C->y);
}

static double ProxDist (SparseTriT * st, PointT * A, PointT * B, PointT * C, MagickBooleanType sym)
/* Returns distance Au to Cu where B-Au and B-Cu are unit vectors in directions B-A and B-C.
   Signed, for CW or CCW.

   Distance Au to Cu is 2.0 if A is 180 degrees around B from C,
   or zero if BA coincides with BC.
*/
{
  PointT Au, Cu;
  UnitVect (B, A, &Au);  // Could optimize: this is repeated around each vertex.
  UnitVect (B, C, &Cu);
  CoordT len = PtsDist (&Au, &Cu);

  if (len == 0) {
    len = -TRI_EPSILON;
  } else if (len >= 2) {
    len = 2;
  } else {

    if (AreaX2 (A, B, C) > 0) len = -len;
  }

  return len;
}

static MagickBooleanType AddPoint (SparseTriT * st, CoordT x, CoordT y)
// Returns whether OK.
{
  if (st->NumPts >= st->MaxNumPts) return MagickFalse;
  PointT * pt = &st->points[st->NumPts];
  pt->x = x;
  pt->y = y;
  pt->firstEdge = NULL;
  pt->visited = MagickFalse;

  st->NumPts++;

  if (st->NumPts == 1) {
    st->minX = x;
    st->minY = y;
    st->maxX = x;
    st->maxY = y;
  } else {
    if (st->minX > x) st->minX = x;
    if (st->minY > y) st->minY = y;
    if (st->maxX < x) st->maxX = x;
    if (st->maxY < y) st->maxY = y;
  }

  return MagickTrue;
}

static inline EdgeT * NextEdge (EdgeT * e, PointT * p)
{
  assert (p == e->org || p == e->dst);

  return (p == e->org) ? e->orgNext : e->dstNext;
}

static inline PointT * OtherPoint (EdgeT * e, PointT * p)
{
  assert (p == e->org || p == e->dst);

  return (p == e->org) ? e->dst : e->org;
}

static EdgeT * FindEdgePosn (SparseTriT * st, PointT * p0, PointT * p1, MagickBooleanType sym)
// Finds position within ring of edges around p0 that is correct for new edge p0-p1.
// Returns prev, the edge in the ring after which this is to be inserted.
{
  // Find three edges, not necessarily present or distinct.
  EdgeT 
    *clsCCW = NULL,    // edge CCW to p1 with shortest distance to p1
    *clsCW = NULL,     // edge CW to p1 with shortest distance to p1
    *lastRight = NULL; // last CCW edge, before a CW edge
  double distCCW = 9e9;
  double distCW = -9e9;
  MagickBooleanType FndCcw = MagickFalse;
  MagickBooleanType FndFirstCw = MagickFalse;

  EdgeT * prev = p0->firstEdge;
  EdgeT * next = NextEdge (prev, p0);
  EdgeT * first = next;

  for (;;) {
    assert (p0 == next->org || p0 == next->dst);

    //printf (" fep %i ", OtherPoint (next, p0) - st->points);
    double pd = ProxDist (st, p1,  p0, OtherPoint (next, p0),  sym);

    if (pd > 0) {
      // We have a CCW edge.
      FndCcw = MagickTrue;
      if (distCCW > pd) {
        distCCW = pd;
        clsCCW = next;
      }
    } else { // pd <= 0
      // We have a CW edge.
      if (FndCcw && !FndFirstCw) {
        FndFirstCw = MagickTrue;
        lastRight = prev;
      }
      if (distCW < pd) {
        distCW = pd;
        clsCW = prev;
        //printf ("distCW=%g ", distCW);
      }
    }

    prev = next;
    next = NextEdge (next, p0);
    if (next == first) {
      // We sometimes need to go round twice.
      if (!(!lastRight && clsCCW && clsCW)) {
        break;
      }
    }
  }

  prev = NULL;

  if (lastRight)   prev = lastRight;
  else if (clsCCW) prev = clsCCW;
  else if (clsCW)  prev = clsCW;
  else fprintf (stderr, "  FindEdgePosn: barf\n");

  assert (prev);

  //printf ("  FindEdgePosn prev=%i\n", prev - st->edges);

  return prev;
}

static MagickBooleanType AddEdge (SparseTriT * st, PointT * p0, PointT * p1)
// Returns whether OK.
{
  assert (p0 != p1);

  if (st->NumEdges >= st->MaxNumEdges) {
    fprintf (stderr, "AddEdge bust\n");
    return MagickFalse;
  }
  EdgeT * e = &st->edges[st->NumEdges];
  //printf ("AddEdge %i points %i %i\n", st->NumEdges, p0 - st->points, p1 - st->points);
  e->org = p0;
  e->dst = p1;
  e->orgNext = e->dstNext = e;
  st->NumEdges++;

  p0->visited = MagickTrue;

  if (!p0->firstEdge) p0->firstEdge = e;
  else {

    //printf (" Nextp0 p0=%i firstEdge=%i ", p0 - st->points, p0->firstEdge - st->edges);
    EdgeT * next = p0->firstEdge;
    //EdgeT * prev = next;

    EdgeT * prev = FindEdgePosn (st, p0, p1, p0==next->dst);

    // Insert new edge after "prev".
    //printf (" prev=%i  New=%i\n", prev - st->edges, e - st->edges);

    if (prev->org == p0) {
      EdgeT * t = prev->orgNext;
      prev->orgNext = e;
      e->orgNext = t;
    } else {
      EdgeT * t = prev->dstNext;
      prev->dstNext = e;
      e->orgNext = t;
    }
  }

  if (!p1->firstEdge) p1->firstEdge = e;
  else {
    //printf (" Nextp1 p1=%i firstEdge=%i ", p1 - st->points, p1->firstEdge - st->edges);
    EdgeT * next = p1->firstEdge;
    //EdgeT * prev = next;

    EdgeT * prev = FindEdgePosn (st, p1, p0, p1==next->dst);

    // Insert new edge after "prev".
    //printf (" prev=%i  New=%i\n", prev - st->edges, e - st->edges);

    if (prev->org == p1) {
      EdgeT * t = prev->orgNext;
      prev->orgNext = e;
      e->dstNext = t;
    } else {
      EdgeT * t = prev->dstNext;
      prev->dstNext = e;
      e->dstNext = t;
    }

  }

  //WrVertexEdges (st, p0);
  //WrVertexEdges (st, p1);

  return MagickTrue;
}

static int comp_ce (const void *a, const void *b)
{
  CoordT sda = ((const CandidateEdgeT *)a)->SqrDist;
  CoordT sdb = ((const CandidateEdgeT *)b)->SqrDist;

  // Sorting also by x and y isn't necessry,
  // but helps when visually examining diagrams.

  if (sda == sdb) {
    sda = ((const CandidateEdgeT *)a)->pa->x;
    sdb = ((const CandidateEdgeT *)b)->pb->x;
  }

  if (sda == sdb) {
    sda = ((const CandidateEdgeT *)a)->pa->y;
    sdb = ((const CandidateEdgeT *)b)->pb->y;
  }

  return (sda > sdb) - (sda < sdb);
}

static MagickBooleanType TriangulateSrtEdge (SparseTriT * st)
// Returns whether triangulation was possible.
{
  // Build list of candidate edges.
  //
  int NumCand = st->NumPts * (st->NumPts-1) / 2;

  if (NumCand <= 0) return MagickFalse;

  CandidateEdgeT * ces = (CandidateEdgeT *)malloc (NumCand * sizeof (CandidateEdgeT));
  if (!ces) {
    fprintf (stderr, "oom ces");
    exit (1);
  }

  int i, j;
  CandidateEdgeT * ce = ces;
  for (i = 0; i < st->NumPts-1; i++) {
    PointT * pti = &st->points[i];
    for (j = i+1; j < st->NumPts; j++) {
      PointT * ptj = &st->points[j];
      ce->pa = pti;
      ce->pb = ptj;
      CoordT dx = ptj->x - pti->x;
      CoordT dy = ptj->y - pti->y;
      ce->SqrDist = dx*dx + dy*dy;
if (ce->SqrDist == 0) printf ("ce->SqrDist==0\n");
      assert (ce->SqrDist > TRI_EPSILON);
      ce++;
    }
  }

  // Sort list of candidate edges.
  //
  qsort ((void *)ces, NumCand, sizeof (CandidateEdgeT), comp_ce);

  // For each candidate edge, are the ends hidden by an existing edge?
  // If not, then add candidate edge as an actual edge.
  //
  int n, ie;
  ce = ces;
  for (n = 0; n < NumCand; n++) {
    MagickBooleanType hidden = MagickFalse;
    for (ie = 0; ie < st->NumEdges; ie++) {
      EdgeT * e = &st->edges[ie];
      hidden |= IsHidden (ce->pa, ce->pb, e->org, e->dst);
      if (hidden) break;
    }
    if (!hidden) {
      if (!AddEdge (st, ce->pa, ce->pb)) return MagickFalse;
    }
    ce++;
  }

  free (ces);

  return MagickTrue;
}

static MagickBooleanType AddTriangles (SparseTriT * st)
// Find the triangles.
// Each triangle has vertices listed in clockwise order.
//   (Strictly speaking, each face is listed with vertices in the order such that the face
//    is on the right of each edge. The algorithm would also find the convex hull,
//    where the face is outside all the points.)
// This would give three listings per triangle,
// so we choose the one with lowest first vertex.
{
  printf ("Triangles:\n");

  printf ("T=%i\n", st->NumEdges - st->NumPts + 2);

  if (st->NumPts < 3) return MagickFalse;

  int i;
  for (i = 0; i < st->NumEdges; i++) {
    EdgeT * e = &st->edges[i];
    e->tri = NULL;
    e->numTri = 0;
  }

  st->MaxNumTris = st->NumEdges - st->NumPts + 4;
  st->NumTris = 0;
  st->tris = (TriangleT *) malloc (st->MaxNumTris * sizeof (TriangleT));
  if (!st->tris) {
    fprintf (stderr, "oom tris");
    return MagickFalse;
  }

  int CntTri = 0;
  for (i = 0; i < st->NumPts; i++) {
    PointT * p0 = &st->points[i];
    EdgeT * e0 = p0->firstEdge;
    do {
      PointT * p1 = OtherPoint (e0, p0);
      //printf ("\np0,p1: %i %i  ", p0 - st->points, p1 - st->points);
      if (p0 < p1) {
        EdgeT * e1 = NextEdge (e0, p1);
        PointT * p2 = OtherPoint (e1, p1);
        EdgeT * e2 = NextEdge (e1, p2);
        PointT * p3 = OtherPoint (e2, p2);

        if (p0 < p2) {
          if (p3 == p0 && AreaX2 (p0, p1, p2) > 0) {
            printf ("  triPts: %li %li %li  # %i\n",
              (long int)(p0 - st->points),
              (long int)(p1 - st->points),
              (long int)(p2 - st->points), CntTri);
            printf ("  tri: %g %g  %g %g  %g %g",
              p0->x, p0->y, p1->x, p1->y, p2->x, p2->y);

            printf ("  A=%.*g\n", st->precision, AreaX2 (p0, p1, p2)/2);

            if (st->NumTris >= st->MaxNumTris) {
              fprintf (stderr, "AddTriangles bust\n");
              return MagickFalse;
            }
            TriangleT * ptri = &st->tris[st->NumTris];
            ptri->p0 = p0;
            ptri->p1 = p1;
            ptri->p2 = p2;

            ptri->denominator = 
              (p1->y - p2->y)*(p0->x - p2->x)
            + (p2->x - p1->x)*(p0->y - p2->y);

            assert (ptri->denominator != 0);

            ptri->centX = (p0->x + p1->x + p2->x) / 3;
            ptri->centY = (p0->y + p1->y + p2->y) / 3;

            e0->tri = ptri;
            e1->tri = ptri;
            e2->tri = ptri;
            e0->numTri++;
            e1->numTri++;
            e2->numTri++;

            st->NumTris++;
            CntTri++;
          } else {
            printf ("AddTriangles p0!=p3 or neg area ");
            printf ("  dt pts %li %li %li %li\n",
              (long int)(p0 - st->points),
              (long int)(p1 - st->points),
              (long int)(p2 - st->points),
              (long int)(p3 - st->points));
          }
        }
      }
      e0 = NextEdge (e0, p0);
    } while (e0 != p0->firstEdge);
  }
  printf ("CntTri: %i\n", CntTri);

  if (st->NumTris) {
    for (i = 0; i < st->NumEdges; i++) {
      EdgeT * e = &st->edges[i];
      assert (e->tri != NULL);
      assert (e->numTri);
    }
  }

  return MagickTrue;
}

static void DumpPoints (SparseTriT * st, FILE * fh)
{
  fprintf (fh, "NumPoints: %i\n", st->NumPts);

  int i;
  int CntEdge = 0;
  for (i = 0; i < st->NumPts; i++) {
    PointT * pt = &st->points[i];
    fprintf (fh, "  point: %.*g %.*g  # %i\n",
      st->precision, pt->x, st->precision, pt->y, i);

    if (pt->firstEdge) {
      EdgeT * e = pt->firstEdge;
      fprintf (fh, "    first: %li  pts: %li %li\n",
        (long int)(e - st->edges),
        (long int)(e->org - st->points),
        (long int)(e->dst - st->points));
      CntEdge++;
      int Limit = 20;
      do {
        e = NextEdge (e, pt);
        if (e == pt->firstEdge) break;
        fprintf (fh, "     next: %li  pts: %li %li\n",
          (long int)(e - st->edges),
          (long int)(e->org - st->points),
          (long int)(e->dst - st->points));
        CntEdge++;
        Limit--;
      } while (e != pt->firstEdge && Limit > 0);
      if (Limit == 0) fprintf (fh, "Loop?\n");
    }
  }
  fprintf (fh, "CountedEdges: %i\n", CntEdge);
}

static void DumpEdges (SparseTriT * st, FILE * fh)
{
  fprintf (fh, "NumEdges: %i\n", st->NumEdges);

  int i;
  for (i = 0; i < st->NumEdges; i++) {
    EdgeT * e = &st->edges[i];
    fprintf (fh, "  edge: %.*g %.*g  %.*g %.*g  # %i  numTri: %i",
      st->precision, e->org->x, st->precision, e->org->y,
      st->precision, e->dst->x, st->precision, e->dst->y,
      i,
      e->numTri);
    if (e->numTri) {
      fprintf (fh, "  tri: %li",
        (long int)(e->tri - st->tris));
    } else {
      fprintf (fh, "  tri: null");
    }
    fprintf (fh, "\n");
  }
}

static void DumpTriangles (SparseTriT * st, FILE * fh)
{
  fprintf (fh, "NumTriangles: %i\n", st->NumTris);
  int i;
  for (i = 0; i < st->NumTris; i++) {
    TriangleT * tri = &st->tris[i];
    fprintf (fh, "  trianglePts: %li %li %li  # %i\n",
      (long int)(tri->p0 - st->points),
      (long int)(tri->p1 - st->points),
      (long int)(tri->p2 - st->points),
      i);
    fprintf (fh, "  triangle: %.*g %.*g  %.*g %.*g  %.*g %.*g\n",
      st->precision, tri->p0->x, st->precision, tri->p0->y,
      st->precision, tri->p1->x, st->precision, tri->p1->y,
      st->precision, tri->p2->x, st->precision, tri->p2->y);
    fprintf (fh, "  triangleCentXY: %.*g %.*g\n",
      st->precision, tri->centX, st->precision, tri->centY);
  }
}

static void DeInitTri (SparseTriT * st)
{
  if (st->MaxNumPts) free (st->points);
  if (st->MaxNumEdges) free (st->edges);
  st->NumPoints = 0;
  st->MaxNumPts = 0;
  st->NumPts = 0;
  st->MaxNumEdges = 0;
  st->NumEdges = 0;

  if (st->MaxNumTris > 0) free (st->tris);
  st->MaxNumTris = 0;
  st->NumTris = 0;
}

static MagickBooleanType InitTri (SparseTriT * st)
/* Args is address of first element of an array with (NumPoints*NumArgsPerPnt) values.
   Within each group of NumArgsPerPnt values, first two are x,y coords.
   Others are values for the barycentric weighting.
   Returns whether OK. If false, no need to call DeInitTri.
*/
{
  if (st->NumPoints < 1) return MagickFalse;
  if (st->NumArgsPp < 2) return MagickFalse;

  st->MaxNumPts = st->NumPoints;
  st->NumPts = 0;
  st->points = (PointT *) malloc (st->MaxNumPts * sizeof (PointT));
  if (!st->points) {
    st->MaxNumPts = 0;
    return MagickFalse;
  }

  st->MaxNumEdges = st->NumPoints * 3;
  st->NumEdges = 0;
  st->edges = (EdgeT *) malloc (st->MaxNumEdges * sizeof (EdgeT));
  if (!st->edges) {
    DeInitTri (st);
    return MagickFalse;
  };

  st->MaxNumTris = 0;
  st->NumTris = 0;
  st->tris = NULL;

  int CoordOffset = st->CoordsLast ? st->NumArgsPp-2 : 0;
  int DataOffset  = st->CoordsLast ? 0 : 2;
  CoordOffset = 0;
  DataOffset = 2;

  int i;
  for (i = 0; i < st->NumPoints; i++) {
    if (!AddPoint (st, 
                   st->pArgs[i*st->NumArgsPp+CoordOffset],
                   st->pArgs[i*st->NumArgsPp+CoordOffset+1]))
    {
      DeInitTri (st);
      return MagickFalse;
    }
    if (st->NumArgsPp <=2)
      st->points[i].pArgs = NULL;
    else
      st->points[i].pArgs = &st->pArgs[i*st->NumArgsPp+DataOffset];
  }

  if (!TriangulateSrtEdge (st)) {
    DeInitTri (st);
    return MagickFalse;
  }

  AddTriangles (st);

  if (st->verbose) {
    fprintf (stderr, "MinCoord: %.*g %.*g\n",
      st->precision, st->minX, st->precision, st->minY);
    fprintf (stderr, "MaxCoord: %.*g %.*g\n",
      st->precision, st->maxX, st->precision, st->maxY);
    DumpPoints (st, stderr);
    DumpEdges (st, stderr);
    DumpTriangles (st, stderr);
  }

  return MagickTrue;
}

/* ---------------------------------------------------------------
   End of triangulation functions.

   Next function: find the triangle for a point.
-----------------------------------------------------------------*/

static CoordT inline DistanceZeroOne (CoordT a)
{
  return (a < 0) ? -a : ( (a >= 1) ? a-1 : 0 );
}

static CoordT inline DistanceZero (CoordT a)
{
  return (a < 0) ? -a : 0;
}

static MagickBooleanType inline WithinZeroOne (CoordT a)
{
  return (DistanceZeroOne(a) < TRI_EPSILON);
}

static void inline Clamp01 (double * a)
{
  *a = (*a <= 0) ? 0 : ( (*a >= 1) ? 1 : *a );
}

/*---
static double PerpDist (SparseTriT * st, PointT * A, PointT * B, PointT * C)
// Retuns perpendicular distance from A to infinite line defined by B and C.
// Assumes B and C are distinct.
{
  double dx = C->x - B->x;
  double dy = C->y - B->y;

  double lenSqr = dx*dx + dy*dy;

  assert (lenSqr != 0);

  double area = fabs(AreaX2 (A, B, C))/2;

  return area / sqrt(lenSqr);
}
---*/

static double ShortestDist (SparseTriT * st, PointT * A, PointT * B, PointT * C, int Debug)
// Finds shortest distance between A and segment BC.
// This is either distance AB or distance BC
// or (where perp dropped from A to line BC is within BC) the perp distance.
{

  // Where AB is line and C is point:
  //t=[(Cx-Ax)(Bx-Ax)+(Cy-Ay)(By-Ay)]/[(Bx-Ax)^2+(By-Ay)^2]

  // Where BC is line and A is point:
  // FIXME: efficiency: pre-calculate edge squared lengths.
  double dx = C->x - B->x;
  double dy = C->y - B->y;

  //if (Debug) printf ("dx=%g dy=%g numer=%g denom=%g  ",
  //  dx, dy, ((A->x - B->x)*dx + (A->y - B->y)*dy), (dx*dx+dy*dy));

  double t = ((A->x - B->x)*dx + (A->y - B->y)*dy)
           / (dx*dx + dy*dy);

  // t is parametric distance along infinite line defined by BC of perpendicular dropped from A.
  // eg if dx==0, t = (A->y - B->y)/dy

  //N->x=Ax+t(Bx-Ax)
  //N->y=Ay+t(By-Ay)

  if (Debug) printf ("t=%g ", t);

  Clamp01 (&t);

  double fndX = B->x + t * (C->x-B->x);
  double fndY = B->y + t * (C->y-B->y);

if (Debug) {
  printf ("fndXY %g %g  ", fndX, fndY);
}

  dx = fndX - A->x;
  dy = fndY - A->y;
  return sqrt (dx*dx + dy*dy);  // FIXME: efficiency: we don't need to sqrt it.
}

static void inline CalcAbc (SparseTriT * st, CoordT x, CoordT y,
  TriangleT * t,
  CoordT * a, CoordT * b, CoordT * c, MagickBooleanType Normalise)
{
  Normalise = MagickTrue;

  *a = ((t->p1->y - t->p2->y)*(x - t->p2->x) + (t->p2->x - t->p1->x)*(y - t->p2->y))
       / t->denominator;
  *b = ((t->p2->y - t->p0->y)*(x - t->p2->x) + (t->p0->x - t->p2->x)*(y - t->p2->y))
       / t->denominator;
  *c = 1 - *a - *b;

  if (!Normalise) {
    *a *= t->denominator;
    *b *= t->denominator;
    *c *= t->denominator;
  }
}

static CoordT inline WorstAbc (SparseTriT * st, CoordT x, CoordT y,
  TriangleT * t)
{
  CoordT a, b, c;

  CalcAbc (st, x, y, t, &a, &b, &c, MagickFalse);

  // For non-normalised, just test against zero.
  a = DistanceZeroOne (a);
  b = DistanceZeroOne (b);
  c = DistanceZeroOne (c);

  if (b > a) a = b;
  if (c > a) a = c;

  return a;
}


static CoordT inline CosAng (PointT * A, PointT * B, PointT * C)
// Returns cosine of angle BAC.
// CosBAC = (b^2 + c^2 - a^2) / (2.b.c)
{
  CoordT a2 = PtsDistSqr (B, C);
  CoordT b2 = PtsDistSqr (C, A);
  CoordT c2 = PtsDistSqr (A, B);
  return (b2 + c2 - a2) / (2 * sqrt(b2) * sqrt(c2));
}

//#define DEBUG_COND (x < 10 && y < 10)
//#define DEBUG_COND ((x > 170 && x < 180 && y > 190) || (x > 188 && x < 202 && y > 196 && y < 201))
//#define DEBUG_COND (x > 188 && x < 202 && y > 43 && y < 52)
//#define DEBUG_COND (x > 130 && x < 135 && y > 190)
//#define DEBUG_COND (x > 138 && x < 146 && y > 90 && y < 97)
//#define DEBUG_COND (x > 130 && x < 150 && y > 80 && y < 100)
#define DEBUG_COND (1==0)

static EdgeT * NearestEdge (SparseTriT * st, CoordT x, CoordT y)
// Finds nearest edge on the convex hull.
{
  PointT A;
  A.x = x;
  A.y = y;
  int i;
  double v;
  double leastV = 9e9;
  EdgeT * fndE1 = NULL;
  EdgeT * fndE2 = NULL;
  for (i = 0; i < st->NumEdges; i++) {
    EdgeT * e = &st->edges[i];
    if (e->numTri == 1) {
      v = ShortestDist (st, &A, e->org, e->dst, DEBUG_COND);

if (DEBUG_COND) {
      printf ("Edge=%i  v=%g\n", i, v);
}

      if (leastV > v) {
        leastV = v;
        fndE1 = e;
      }
    }
  }

  assert (fndE1);

if (DEBUG_COND) {
  printf ("fndE=%li  \n", (long int)(fndE1 - st->edges));
}

  // Do we have another edge at practically the same distance?

  for (i = 0; i < st->NumEdges; i++) {
    EdgeT * e = &st->edges[i];
    if (e->numTri == 1 && e != fndE1) {
      v = ShortestDist (st, &A, e->org, e->dst, DEBUG_COND);
      if (fabs (leastV - v) < TRI_EPSILON) {
        fndE2 = e;
      }
    }
  }
  // If two edges are equal, they probably have a common point which we are close to.
  // Choose the edge that gives the largest angle (e->p0, {x,y}, e->p1).

  if (fndE2) {
    CoordT cos1 = CosAng (&A, fndE1->org, fndE1->dst);
    CoordT cos2 = CosAng (&A, fndE2->org, fndE2->dst);

if (DEBUG_COND) {
    printf ("fndE2=%li  cos=%g %g  \n",
      (long int)(fndE2 - st->edges),
      cos1, cos2);
}
    if (cos2 < cos1) fndE1 = fndE2;

  }

  return fndE1;
}

static MagickBooleanType FindTriangle
  (SparseTriParaT * stp, CoordT x, CoordT y,
   TriangleT ** tri,
   CoordT * a, CoordT * b, CoordT * c)
/* Returns whether this coord is within convex hull of triangulated points. */
/* If there are no triangles, don't call this. */
{
  // FIXME: probably worth checking for st->NumTris==1.

  int i;
  MagickBooleanType FoundExact = MagickFalse;

  for (i = 0; i < stp->st->NumTris; i++) {
    int TriNum = (i+stp->PrevFndTri) % stp->st->NumTris;
    TriangleT * t = &stp->st->tris[TriNum];
    *tri = t;

    CalcAbc (stp->st, x, y, t, a, b, c, MagickTrue);

if (DEBUG_COND) {
    printf ("xy=%g %g  PrevTri=%i  Tri=%i abc=%g %g %g   \n",
      x, y, stp->PrevFndTri, TriNum, *a, *b, *c);
}

    if (WithinZeroOne (*a) && WithinZeroOne (*b) && WithinZeroOne (*c)) {
      stp->PrevFndTri = TriNum;
      FoundExact = MagickTrue;
      break;
    }

/*---
    CoordT d = DistanceZeroOne (*a);
    if (d > 0 && d < SmallestN01Dist) {
      SmallestN01Dist = d;
      ClosestTri = TriNum;
    }
    d = DistanceZeroOne (*b);
    if (d > 0 && d < SmallestN01Dist) {
      SmallestN01Dist = d;
      ClosestTri = TriNum;
    }
    d = DistanceZeroOne (*c);
    if (d > 0 && d < SmallestN01Dist) {
      SmallestN01Dist = d;
      ClosestTri = TriNum;
    }
---*/
  }

  if (!FoundExact && stp->st->outside != oBackground) {

    EdgeT * e = NearestEdge (stp->st, x, y);
    *tri = e->tri;

#if 0
    // Find the least worst triangle.

    CoordT LeastV = 99e9;

    for (i = 0; i < st->NumTris; i++) {
      int TriNum = (i+st->PrevFndTri) % st->NumTris;
      TriangleT * t = &st->tris[TriNum];

      CoordT v = WorstAbc (st, x, y, t);
      if (LeastV > v) {
        LeastV = v;
        st->PrevFndTri = TriNum;
        *tri = t;
      }
if (DEBUG_COND) {
  printf ("Tri %i Worst=%g, LeastV=%g  \n", TriNum, v, LeastV);
}

/*---
        double dx = x - t->centX;
        double dy = y - t->centY;
        double distSq = dx*dx + dy*dy;

if (DEBUG_COND) {
  printf ("DistSq=%g  \n", distSq);
}
        if (LeastV > distSq) {
          LeastV = distSq;
          st->PrevFndTri = TriNum;
          *tri = t;
        }
---*/

    }

if (DEBUG_COND) {
    printf ("LeastV=%g ClosestTri=%li  \n", LeastV, (long int)(*tri - st->tris));
}

#endif
if (DEBUG_COND) {
    printf ("ClosestTri=%li  \n", (long int)(*tri - stp->st->tris));
}

    CalcAbc (stp->st, x, y, *tri, a, b, c, MagickFalse);
  }
  return FoundExact;
}



/* ---------------------------------------------------------------

Sigmoidal processing, copied from enhance.c

     a is contrast
     b is midpoint, 0.0 <= b <= 1.0
     x is the value to be converted, 0.0 <= x <= 1.0.

--------------------------------------------------------------- */

#if defined(MAGICKCORE_HAVE_ATANH)
#define Sigmoidal(a,b,x) ( tanh((0.5*(a))*((x)-(b))) )
#else
#define Sigmoidal(a,b,x) ( 1.0/(1.0+exp((a)*((b)-(x)))) )
#endif

#define ScaledSigmoidal(a,b,x) (                    \
  (Sigmoidal((a),(b),(x))-Sigmoidal((a),(b),0.0)) / \
  (Sigmoidal((a),(b),1.0)-Sigmoidal((a),(b),0.0)) )

static inline double InverseScaledSigmoidal(const double a,const double b,
  const double x)
{
  const double sig0=Sigmoidal(a,b,0.0);
  const double sig1=Sigmoidal(a,b,1.0);
  const double argument=(sig1-sig0)*x+sig0;
  const double clamped=
    (
#if defined(MAGICKCORE_HAVE_ATANH)
      argument < -1+MagickEpsilon
      ?
      -1+MagickEpsilon
      :
      ( argument > 1-MagickEpsilon ? 1-MagickEpsilon : argument )
    );
  return(b+(2.0/a)*atanh(clamped));
#else
      argument < MagickEpsilon
      ?
      MagickEpsilon
      :
      ( argument > 1-MagickEpsilon ? 1-MagickEpsilon : argument )
    );
  return(b-log(1.0/clamped-1.0)/a);
#endif
}


/* ---------------------------------------------------------------

Entry points:

--------------------------------------------------------------- */

static inline double AbcPow (double abc, double p)
{
  return (abc < 0) ? -pow(-abc,p) : pow(abc,p);
}

static MagickBooleanType GetTriVals
  (SparseTriParaT * stp, CoordT x, CoordT y, double ** TriRetVals)
/* TriRetVals must be an array already allocated (and eventally freed)
   by the caller with at least (NumArgsPp-2) entries. */
{
  if (!stp->st->NumTris) return MagickFalse;

  CoordT a, b, c;

  TriangleT * tri = NULL;

if (DEBUG_COND) {
  printf ("xy=%g %g   ", x, y);
}

  MagickBooleanType FndExact = FindTriangle (stp, x, y, &tri, &a, &b, &c);

  if (FndExact || stp->st->outside != oBackground) {
if (DEBUG_COND) {
    printf ("Exact=%i  tri=%li  abc=%g %g %g   ",
      FndExact, (long int)(tri - stp->st->tris), a, b, c);
}

    if (stp->st->power != 1.0) {
      a = AbcPow (a, stp->st->power);
      b = AbcPow (b, stp->st->power);
      c = AbcPow (c, stp->st->power);
    }

    if (stp->st->sigContrast > 0.0) {
      a = ScaledSigmoidal(stp->st->sigContrast,stp->st->sigMidpoint,a);
      b = ScaledSigmoidal(stp->st->sigContrast,stp->st->sigMidpoint,b);
      c = ScaledSigmoidal(stp->st->sigContrast,stp->st->sigMidpoint,c);
    } else if (stp->st->sigContrast < 0.0) {
      a = InverseScaledSigmoidal(-stp->st->sigContrast,stp->st->sigMidpoint,a);
      b = InverseScaledSigmoidal(-stp->st->sigContrast,stp->st->sigMidpoint,b);
      c = InverseScaledSigmoidal(-stp->st->sigContrast,stp->st->sigMidpoint,c);
    }

    if (stp->st->outside == oClamp) {
      Clamp01 (&a);
      Clamp01 (&b);
      Clamp01 (&c);
    }

    double div = a + b + c;

    int i;
    for (i = 0; i < stp->st->NumArgsPp-2; i++) {
      (*TriRetVals)[i] = (tri->p0->pArgs[i] * a
                        + tri->p1->pArgs[i] * b
                        + tri->p2->pArgs[i] * c) / div;

if (DEBUG_COND) {
      printf ("RV[%i]=%g ", i, (*TriRetVals)[i]);
}

    }

if (DEBUG_COND) {
    printf ("\n");
}

  }
  return FndExact;
}

